#import "PGPfrontEnd.h"

@implementation PGPfrontEnd

- init
{
    [super init];
    cryptor = [[PGPserver alloc] init];
    keyRingArray = nil;

    [outputTextView setRichText:YES];

    return self;
}


- (void)dealloc
{
    [cryptor release];
    [super dealloc];
    return;
}


- encrypt:(NSPasteboard *)pboard userData:(NSString *)inData error:(NSString **)error
{
    NSMutableData *pgpIn,*pgpOut;
    NSString *pgpOutString;
    int irow;
    NSString *encryptionID;
    NSMutableArray *userIDArray;
    BOOL doSign,doClearSign,doDetachSig;
    BOOL doConventional,doEncryptToSelf;
    BOOL doAsciiArmor;
    NSString *passphrase;
    NSArray *types;
    int pgperror;

    pgpIn = [[self getPBStringData:pboard] retain];
    if(!pgpIn) return nil;

    /* Load tableView with userIDs from keyring */
    if(keyRingArray) [keyRingArray release];
    keyRingArray = [[cryptor getAllUserIDs:&pgperror] retain];
    [userIDTableView reloadData];

    /* Select options and userID from the tableView */
    [selectKeyPanel makeKeyAndOrderFront:self];
    if([NSApp runModalForWindow:selectKeyPanel]==0) {
        *error = nil;
        return nil;
    }

    /* Get options */
    doAsciiArmor = [[optionButtons cellAtRow:0 column:0] state];
    doSign = [[optionButtons cellAtRow:1 column:0] state];
    doClearSign = [[optionButtons cellAtRow:2 column:0] state];
    doDetachSig = [[optionButtons cellAtRow:3 column:0] state];
    doConventional = [[optionButtons cellAtRow:4 column:0] state];
    doEncryptToSelf = [[optionButtons cellAtRow:5 column:0] state];

    [cryptor doArmor:doAsciiArmor];
    [cryptor doConventionalOnly:doConventional];
    [cryptor doSignature:doSign];
    [cryptor clearSign:doClearSign];
    [cryptor detachSignature:doDetachSig];
    [cryptor encryptToSelf:doEncryptToSelf];

    /* Extract all selected userIDs */
    userIDArray = [[NSMutableArray alloc] init];
    for(irow=0; irow<[userIDTableView numberOfRows]; irow++) {
	if([userIDTableView isRowSelected:irow]==YES) {
	    encryptionID = [keyRingArray objectAtIndex:irow];
	    [userIDArray addObject:encryptionID];
	}
    }

    if([userIDArray count]==0) {
	[userIDArray release];
	userIDArray = nil;
    }

    if(doSign || doClearSign || doConventional) {
        /* Get a passphrase */
        [passphrasePanel makeKeyAndOrderFront:self];
	[passphrasePanel makeFirstResponder:passphraseText];
        if([NSApp runModalForWindow:passphrasePanel] == 0) return nil;

        passphrase = [[passphraseText stringValue] retain];
    }

    pgpOut = [[cryptor encryptData:pgpIn
                          withKeys:userIDArray
                        passPhrase:passphrase
                             error:&pgperror] retain];

    /* Check for errors */
    if(pgperror != 0) {
	NSRunAlertPanel(@"PGP error",@"Encountered PGP error %d while encrypting",
			0,0,0,pgperror);
	[pgpIn release];
        *error = nil;
	return nil;
    }

    if([pgpOut length] == 0) {
	NSLog(@"PGP error resulted in null output for selection");
	[pgpIn release];
        *error = nil;
	return nil;
    }

    /* Convert back to NSString */
    pgpOutString = [[NSString alloc] initWithData:pgpOut encoding:NSASCIIStringEncoding];

    types = [NSArray arrayWithObject:NSStringPboardType];
    [pboard declareTypes:types owner:self];
    [pboard setString:pgpOutString forType:NSStringPboardType];

    [self showStatus:self];
    [pgpIn release];

    *error = nil;
    return self;
}

- acceptUserID:sender
{
    [NSApp stopModalWithCode:1];
    [selectKeyPanel orderOut:self];
    return self;
}

- cancelUserID:sender
{
    [NSApp stopModalWithCode:0];
    [selectKeyPanel orderOut:self];
    return self;
}

- acceptPassphrase:sender
{
    [NSApp stopModalWithCode:1];
    [passphrasePanel orderOut:self];
    return self;
}


- cancelPassphrase:sender
{
    [NSApp stopModalWithCode:0];
    [passphrasePanel orderOut:self];
    return self;
}




- decrypt:(NSPasteboard *)pboard userData:(NSString *)inData error:(NSString **)error
{
    NSArray *types;
    NSString *passPhrase;
    NSData *pgpIn,*pgpOut;
    int pgperror;
    NSString *pgpOutString;

    pgpIn = [[self getPBStringData:pboard] retain];
    if(!pgpIn) return nil;

    /* Get a passphrase */
    [passphrasePanel makeKeyAndOrderFront:self];
    [passphrasePanel makeFirstResponder:passphraseText];
    if([NSApp runModalForWindow:passphrasePanel] == 0) return nil;

    passPhrase = [[passphraseText stringValue] retain];

    pgpOut = [cryptor decryptData:pgpIn passPhrase:passPhrase error:&pgperror];

    if(pgperror != 0) {
        switch (pgperror) {
	  case -11500:
	    NSRunAlertPanel(@"Decrypt PGP Error",@"Bad Passphrase",
			    0,0,0);
	    break;

	  default:

	    NSRunAlertPanel(@"Decrypt PGP Error",@"Encountered PGP error %d while decrypting",
			    0,0,0,pgperror);
	    break;
	}

        [pgpIn release];
	return nil;
    }

    if([pgpOut length] == 0) {
	NSLog(@"PGP error resulted in null output for selection");
        [pgpIn release];
	return nil;
    }

    /* Convert back to NSString */
    pgpOutString = [[NSString alloc] initWithData:pgpOut encoding:NSASCIIStringEncoding];

    types = [NSArray arrayWithObject:NSStringPboardType];
    [pboard declareTypes:types owner:self];
    [pboard setString:pgpOutString forType:NSStringPboardType];

    [pgpIn release];
    [self showStatus:self];

    *error = nil;
    return self;
}


- decryptInWindow:(NSPasteboard *)pboard userData:(NSString *)inData error:(NSString **)error
{
    NSString *passPhrase;
    NSData *pgpIn,*pgpOut;
    int pgperror;
    NSRange outputRange;

    pgpIn = [self getPBStringData:pboard];
    if(!pgpIn) return nil;

    /* Get a passphrase */
    [passphrasePanel makeKeyAndOrderFront:self];
    [passphrasePanel makeFirstResponder:passphraseText];
    if([NSApp runModalForWindow:passphrasePanel] == 0) return nil;
    passPhrase = [[passphraseText stringValue] retain];

    pgpOut = [cryptor decryptData:(NSMutableData *)pgpIn passPhrase:passPhrase error:&pgperror];

    if(pgperror) {
	NSRunAlertPanel(@"Decrypt PGP Error",@"Encountered PGP error %d while decrypting",
			0,0,0,pgperror);
	return nil;
    }

    if([pgpOut length] == 0) {
	NSLog(@"PGP error resulted in null output for selection");
	return nil;
    }
    
    outputRange.location = 0;
    outputRange.length = [[outputTextView string] length];

    if(!strncmp((char *)[pgpOut bytes],"{\\rtf",5))
	[outputTextView replaceCharactersInRange:outputRange withRTF:pgpOut];
    else {
	[outputTextView setSelectedRange:outputRange];
	[outputTextView insertText:
			[[[NSString alloc] initWithData:pgpOut encoding:NSASCIIStringEncoding] autorelease]];
    }

    [outputTextView display];
    [self showOutput:self];
    [self showStatus:self];

    *error = nil;
    return self;
}



/*************************   PGP FILE MANAGEMENT ROUTINES HERE **************************/

- encryptFiles:(NSPasteboard *)pboard userData:(NSString *)inData error:(NSString **)error
{
    NSArray *fileArray;
    BOOL doSign,doClearSign,doDetachSig;
    BOOL doConventional,doEncryptToSelf;
    BOOL doAsciiArmor;
    int irow;
    NSString *encryptionID;
    NSMutableArray *userIDArray;
    NSString *passPhrase;
    int pgperror;
    int ifile;

    fileArray = [self getPBFileNames:pboard];

    /* Load tableView with userIDs from keyring */
    if(keyRingArray) [keyRingArray release];
    keyRingArray = [[cryptor getAllUserIDs:&pgperror] retain];
    [userIDTableView reloadData];

    /* Select an encryption userID from the tableView */
    [selectKeyPanel makeKeyAndOrderFront:self];
    if([NSApp runModalForWindow:selectKeyPanel]==0) {
        *error = nil;
        return nil;
    }

    /* Get options */
    doAsciiArmor = [[optionButtons cellAtRow:0 column:0] state];
    doSign = [[optionButtons cellAtRow:1 column:0] state];
    doClearSign = [[optionButtons cellAtRow:2 column:0] state];
    doDetachSig = [[optionButtons cellAtRow:3 column:0] state];
    doConventional = [[optionButtons cellAtRow:4 column:0] state];
    doEncryptToSelf = [[optionButtons cellAtRow:5 column:0] state];

    [cryptor doArmor:doAsciiArmor];
    [cryptor doConventionalOnly:doConventional];
    [cryptor doSignature:doSign];
    [cryptor clearSign:doClearSign];
    [cryptor detachSignature:doDetachSig];
    [cryptor encryptToSelf:doEncryptToSelf];

    /* Extract the selected userID */
    userIDArray = [[NSMutableArray alloc] init];
    for(irow=0; irow<[userIDTableView numberOfRows]; irow++) {
	if([userIDTableView isRowSelected:irow]==YES) {
	    encryptionID = [keyRingArray objectAtIndex:irow];
	    [userIDArray addObject:encryptionID];
	}
    }

    if([userIDArray count]==0) {
	[userIDArray release];
	userIDArray = nil;
    }

    if(doSign || doClearSign || doConventional) {
        /* Get a passphrase */
        [passphrasePanel makeKeyAndOrderFront:self];
	[passphrasePanel makeFirstResponder:passphraseText];
        if([NSApp runModalForWindow:passphrasePanel] == 0) return nil;

        passPhrase = [[passphraseText stringValue] retain];
    }

    [encryptInfoPanel orderFront:self];
    
    for(ifile=0; ifile<[fileArray count]; ifile++) {
        NSAutoreleasePool *pool;
	NSString *oldFile,*newFile;

	pool = [[NSAutoreleasePool alloc] init];
	oldFile = [fileArray objectAtIndex:ifile];
	newFile = [oldFile stringByAppendingPathExtension:@"pgp"];

        [encryptFileTextField setStringValue:oldFile];
        [encryptInfoPanel display];
        [encryptInfoPanel flushWindow];
        [[NSDPSContext currentContext] wait];

	[cryptor encryptFile:oldFile
                      toFile:newFile
                    withKeys:userIDArray
		  passPhrase:passPhrase
                       error:&pgperror];

	/* Check for errors */
	if(pgperror) {
	    NSRunAlertPanel(@"PGP error",@"Encountered PGP error %d while encrypting file %@",
			    0,0,0,pgperror,oldFile);
	    continue;
	}
	[pool release];
    }

    [encryptInfoPanel orderOut:self];
    [self showStatus:self];

    *error = nil;
    return self;
}




- decryptFiles:(NSPasteboard *)pboard userData:(NSString *)inData error:(NSString **)error
{
    NSArray *fileArray;
    NSString *passPhrase;
    int pgperror;
    int ifile;

    *error = nil;
    
    fileArray = [self getPBFileNames:pboard];

    /* Get a passphrase */
    [passphrasePanel makeKeyAndOrderFront:self];
    [passphrasePanel makeFirstResponder:passphraseText];
    if([NSApp runModalForWindow:passphrasePanel] == 0) return nil;

    passPhrase = [[passphraseText stringValue] retain];

    [decryptInfoPanel orderFront:self];

    for(ifile=0; ifile<[fileArray count]; ifile++) {
	NSAutoreleasePool *pool;
	NSString *oldFile,*newFile;

        pool = [[NSAutoreleasePool alloc] init];
	oldFile = [fileArray objectAtIndex:ifile];
	newFile = [oldFile stringByDeletingPathExtension];

        [decryptFileTextField setStringValue:oldFile];
        [decryptInfoPanel display];
        [decryptInfoPanel flushWindow];
        [[NSDPSContext currentContext] wait];

	[cryptor decryptFile:oldFile passPhrase:passPhrase plainFile:newFile error:&pgperror];

	/* Check for errors */
	if(pgperror) {
	    NSRunAlertPanel(@"PGP error",@"Encountered PGP error %d while decrypting file %@",
			    0,0,0,pgperror,oldFile);
	    continue;
	}
    }

    [decryptInfoPanel orderOut:self];

    [self showStatus:self];
    *error = nil;
    return self;
}


- clearSign:(NSPasteboard *)pboard userData:(NSString *)inData error:(NSString **)error
{
    NSData *docData,*pgpOut;
    NSString *pgpOutString;
    NSString *passPhrase;
    NSArray *types;
    int pgperror;

    docData = [self getPBStringData:pboard];
    if(!docData) return nil;
    
    /* Get a passphrase */
    [passphrasePanel makeKeyAndOrderFront:self];
    [passphrasePanel makeFirstResponder:passphraseText];
    if([NSApp runModalForWindow:passphrasePanel] == 0) return nil;

    passPhrase = [[passphraseText stringValue] retain];

    /* Sign the data */
    [cryptor clearSign:YES];
    pgpOut = [cryptor signData:docData
                    passPhrase:passPhrase
                         error:&pgperror];

    if(pgperror) {
	NSRunAlertPanel(@"PGP error",@"Encountered PGP error %d while signing document",
			0,0,0,pgperror);
	return nil;
    }

    /* Convert back to NSString */
    pgpOutString = [[NSString alloc] initWithData:pgpOut encoding:NSASCIIStringEncoding];

    types = [NSArray arrayWithObject:NSStringPboardType];
    [pboard declareTypes:types owner:self];
    [pboard setString:pgpOutString forType:NSStringPboardType];

    [self showStatus:self];

    *error = nil;
    return self;
}


- detachedSig:(NSPasteboard *)pboard userData:(NSString *)inData error:(NSString **)error
{
    NSData *pgpIn,*pgpOut;
    NSString *passPhrase;
    int pgperror;
    NSRange outputRange;

    pgpIn = [self getPBStringData:pboard];
    if(!pgpIn) return nil;

    /* Get a passphrase */
    [passphrasePanel makeKeyAndOrderFront:self];
    [passphrasePanel makeFirstResponder:passphraseText];
    if([NSApp runModalForWindow:passphrasePanel] == 0) return nil;
    passPhrase = [[passphraseText stringValue] retain];

    [cryptor doSignature:YES];
    [cryptor detachSignature:YES];
    pgpOut = [cryptor decryptData:pgpIn passPhrase:passPhrase error:&pgperror];

    if(pgperror) {
	NSRunAlertPanel(@"PGP Signature Error",@"Encountered PGP error %d while signing data",
			0,0,0,pgperror);
	return nil;
    }

    if([pgpOut length] == 0) {
	NSLog(@"PGP error resulted in null output for selection");
	return nil;
    }

    outputRange.location = 0;
    outputRange.length = [[outputTextView string] length];

    [outputTextView setSelectedRange:outputRange];
    [outputTextView insertText:
                        [[[NSString alloc] initWithData:pgpOut encoding:NSASCIIStringEncoding] autorelease]];

    [outputTextView display];
    [self showStatus:self];

    *error = nil;
    return self;
}



- checkSig:(NSPasteboard *)pboard userData:(NSString *)inData error:(NSString **)error
{
    NSData *docData,*result;
    int pgperror,sigStat;
    NSDate *sigDate;
    NSString *sigID;

    docData = [self getPBStringData:pboard];
    result = [cryptor verifySignatureData:docData passPhrase:nil error:&pgperror];
    if(pgperror) {
	NSRunAlertPanel(@"Signature Verify PGP Error",@"Encountered PGP error %d while verifying signature",
			0,0,0,pgperror);
	return nil;
    }

    sigStat = [cryptor signatureStatus];
    switch(sigStat) {
        case PGP6_NOSIG:
            NSRunAlertPanel(@"PGP Signature",@"Document Not Signed",0,0,0);
            break;

        case PGP6_GOODSIG:
            sigDate = [cryptor signatureDate];
            sigID = [cryptor signator];

            NSRunAlertPanel(@"PGP Signature",@"Verified Good Signature by %s on date %s",0,0,0,
                            [sigID cString],[[sigDate description] cString]);
            break;

        case PGP6_SIGNOKEY:
            NSRunAlertPanel(@"PGP Signature",@"Unverified Signature: No public key",0,0,0);
            break;

        case PGP6_BADSIG:
            NSRunAlertPanel(@"PGP Signature",@"Document has Bad Signature",0,0,0);
            break;
    }

    return self;
}


- showStatus:sender
{
    if([[statusTextView string] length] == 0) return self;
    [statusPanel orderFront:self];
    return self;
}


- showOutput:sender
{
    if([[outputTextView string] length] == 0) return self;
    [outputPanel orderFront:self];
    return self;
}



/************************  KEY MANAGEMENT ROUTINES *******************************/


- addKey:(NSPasteboard *)pboard userData:(NSString *)inData error:(NSString **)error
{
    NSData *pgpKey;
    int pgperror;

    pgpKey = [self getPBStringData:pboard];
    if(!pgpKey) return nil;

    [cryptor importKeyData:pgpKey error:&pgperror];

    if(pgperror != 0) {
	NSRunAlertPanel(@"PGP Error",@"Encountered PGP error %d while adding key to keyring",
			0,0,0,pgperror);
	return nil;
    }

    NSRunAlertPanel(@"PGP Add Key",@"Key Added to Keyring Successfully",0,0,0);
	
    return self;
}


- insertKey:(NSPasteboard *)pboard userData:(NSString *)inData error:(NSString **)error
{
    int pgperror;
    NSString *userID;
    NSData *keyData;
    char *idBuf,*newline;
    NSString *keyString;
    NSArray *types;

    /* Get userID from pasteboard */
    userID = [self getPBString:pboard];
    if(!userID) return nil;

    /* Remove carriage returns, etc */
    idBuf = malloc(1+[userID length]);
    strcpy(idBuf,[userID cString]);
    newline = strchr(idBuf,'\n');
    if(newline != NULL) *newline = '\0';

    userID = [[NSString alloc] initWithCString:idBuf];
    free(idBuf);
    
    /* Extract key from keyring */
    keyData = [cryptor exportKeyData:userID  error:&pgperror];
    if(pgperror != 0) {
	NSRunAlertPanel(@"PGP Error",@"Encountered PGP error %d while getting key data for %@",
			0,0,0,pgperror,NSUserName());
	return nil;
    }

    /* Convert back to NSString */
    keyString = [[NSString alloc] initWithData:keyData encoding:NSASCIIStringEncoding];

    types = [NSArray arrayWithObject:NSStringPboardType];
    [pboard declareTypes:types owner:self];
    [pboard setString:keyString forType:NSStringPboardType];

    [userID release];
    [keyString release];
	
    [self showStatus:self];

    return self;
}


- addKeyFile:(NSPasteboard *)pboard userData:(NSString *)inData error:(NSString **)error
{
    NSArray *fileArray;
    int ifile;
    int pgperror;

    fileArray = [self getPBFileNames:pboard];
    if((fileArray == nil) || ([fileArray count]==0)) return nil;

    for(ifile=0; ifile<[fileArray count]; ifile++) {
        [cryptor importKeyFile:[fileArray objectAtIndex:ifile] error:&pgperror];
        if(pgperror != 0) {
            NSRunAlertPanel(@"PGP Error",@"Encountered PGP Error %d while encrypting file %s",
                            0,0,0,pgperror,[[fileArray objectAtIndex:ifile] cString]);
            continue;
        }
    }

    NSRunAlertPanel(@"PGP Add Key",@"PGP Successful adding Keys to keyring",0,0,0);

    return self;
}


/******************    NICE UTILITY METHODS     *************************/

- (NSString *)getPBString:(NSPasteboard *)pboard
{
    NSArray *types;
    NSString *pbIn;

    types = [pboard types];

    if(![types containsObject:NSStringPboardType]) {
        int itype;
        NSLog(@"NXPGP: Pasteboard couldn't give selection");
        NSLog(@"  Pasteboard types are: %d",[types count]);
        for(itype=0; itype<[types count]; itype++)
            NSLog(@"  %s",[[types objectAtIndex:itype] cString]);
        return nil;
    }

    pbIn = [pboard stringForType:NSStringPboardType];
    return pbIn;
}


- (NSMutableData *)getPBStringData:(NSPasteboard *)pboard
{
    NSString *pbIn;
    NSMutableData *pgpIn;

    pbIn = [self getPBString:pboard];
    pgpIn = [NSMutableData dataWithBytes:[pbIn cString] length:[pbIn length]];

    if(!pgpIn) {
	NSLog(@"Can't get string data from pasteboard");
	return nil;
    }

    return pgpIn;
}


- (NSArray *)getPBFileNames:(NSPasteboard *)pboard
{
    NSArray *pasteboardNames;
    NSMutableArray *fileNames;
    NSArray *types;
    int iname;

    types = [pboard types];

    if(![types containsObject:NSFilenamesPboardType]) {
	NSLog(@"NXPGP: Pasteboard couldn't give Filename selection");
	return nil;
    }

    pasteboardNames = [[pboard propertyListForType:NSFilenamesPboardType] copy];
    if(!pasteboardNames) {
	NSLog(@"NXPGP: Pasteboard couldn't give Filename data");
	return nil;
    }

    /* Turn all directory names into an array of all the */
    /* filenames in the subdirectory tree                */

    fileNames = [[[NSMutableArray alloc] init] autorelease];

    for(iname=0; iname<[pasteboardNames count]; iname++) {
	NSString *currentName;
        BOOL isDir;

	isDir = NO;
	currentName = [pasteboardNames objectAtIndex:iname];
        [[NSFileManager defaultManager] fileExistsAtPath:currentName isDirectory:&isDir];
	if(isDir) {
	    NSArray *subDirNames;
	    NSString *fullName;
	    int isubname;

	    subDirNames = [[NSFileManager defaultManager] subpathsAtPath:currentName];
	    for(isubname=0; isubname<[subDirNames count]; isubname++) {
		fullName = [currentName stringByAppendingPathComponent:
			    [subDirNames objectAtIndex:isubname]];

		[fileNames addObject:fullName];
	    }
	}
	else [fileNames addObject:currentName];
    }
    
    return fileNames;
}

    


/******************  NSApplication Delegate Routines *************************/

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification
{
    [NSApp setServicesProvider:self];
    return;
}

#if 0
- (BOOL)application:(NSApplication *)anApp openFile:(NSString *)aFile
{
    NSData *pgpIn,*pgpOut;
    NSMutableArray *options;
    NSString *newFile;
    int result;

    [decryptFileTextField setStringValue:aFile];
    [decryptInfoPanel display];
    [decryptInfoPanel flushWindow];
    [decryptInfoPanel orderFront:self];

    options = [[[NSMutableArray alloc] init] autorelease];
    pgpIn = [[NSData dataWithContentsOfFile:aFile] retain];
    pgpOut = [[cryptor runpgp:DECRYPT options:options data:pgpIn] retain];

    /* Check for errors */
    if(!pgpOut) {
	NSLog(@"NXPGP: PGP cancelled out");
	[decryptInfoPanel orderOut:self];
	[cryptor clearIdeaPhrase:self];
	[cryptor clearDecryptPhrase:self];
	[pgpIn release];
	return NO;
    }

    if([pgpOut length] == 0) {
	NSLog(@"PGP error resulted in null output");
	[decryptInfoPanel orderOut:self];
	[cryptor clearIdeaPhrase:self];
	[cryptor clearDecryptPhrase:self];
#if 0
	[pgpIn release];
#endif
	[pgpOut release];
	return NO;
    }
    
    /* Write back to file */
    newFile = [aFile stringByDeletingPathExtension];
    if([[NSFileManager defaultManager] fileExistsAtPath:newFile]) {
	result = NSRunAlertPanel([NSString stringWithFormat:@"File %@ Exists",newFile],
				 @"Over Write Existing File?",@"Yes",@"No",0);
	switch(result) {
	  case NSAlertDefaultReturn:
	    [pgpOut writeToFile:newFile atomically:NO];
	    break;
	  default:
	    break;
	}
    }
    else
      [pgpOut writeToFile:newFile atomically:NO];

    [pgpIn release];
    [pgpOut release];
    
    [decryptInfoPanel orderOut:self];
    [cryptor clearIdeaPhrase:self];
    [cryptor clearDecryptPhrase:self];

    return YES;
}
#endif




/*************  TableView delegate methods **********************/
- (id)tableView:(NSTableView*)aTableView
                objectValueForTableColumn:(NSTableColumn*)aTableColumn
                row:(int)rowIndex
{
    if(rowIndex > [keyRingArray count]) {
        fprintf(stderr,"Requested object at index %d from array with only %d elements\n",rowIndex,[keyRingArray count]);
        return nil;
    }

    return [keyRingArray objectAtIndex:rowIndex];
}

- (int)numberOfRowsInTableView:(NSTableView *)aTableView
{
    return [keyRingArray count];
}

- (BOOL)tableView:(NSTableView *)aTableView shouldEditTableColumn:(NSTableColumn *)aColumn row:(int)index
{
    return NO;
}


- (void)getKeyFromServer:(NSPasteboard *)pboard userData:(NSString *)inData error:(NSString **)error
{
    NSArray *types;
    NSString *pgpID;
    NSString *newKey;
    NSRange outputRange;
    char *idBuf,*newline;
    int pgperror;

    types = [pboard types];

    if(![types containsObject:NSStringPboardType]) {
	int itype;
	NSLog(@"NXPGP: Pasteboard couldn't give selection");
	NSLog(@"  Pasteboard types are: %d",[types count]);
	for(itype=0; itype<[types count]; itype++)
	    NSLog(@"  %s",[[types objectAtIndex:itype] cString]);
	return;
    }

    /* Get input NSString */
    pgpID = [pboard stringForType:NSStringPboardType];

    /* Remove leading spaces, trailing carriage returns, etc */
    idBuf = malloc(1+[pgpID length]);
    strcpy(idBuf,[pgpID cString]);
    newline = strchr(idBuf,'\n');
    if(newline != NULL) *newline = '\0';
    newline = idBuf;
    while(*newline && (*newline == ' ')) newline++;

    pgpID = [[NSString alloc] initWithCString:newline];
    free(idBuf);
    
    fprintf(stderr,"Getting key for user: |%s|\n",[pgpID cString]);
    
    if(!pgpID) {
	NSLog(@"NXPGP: Pasteboard couldn't give selection data");
	return;
    }

    newKey = [self getKey:pgpID fromServer:@"pgp5.ai.mit.edu" error:&pgperror];

    outputRange.location = 0;
    outputRange.length = [[outputTextView string] length];

    [outputTextView setSelectedRange:outputRange];
    [outputTextView insertText:newKey];

    [outputTextView display];
    [self showOutput:self];
    [self showStatus:self];
    *error = nil;
    return;
}


#define BUFSIZE 128
#define TIMEOUT 10000

- (NSString *)getKey:(NSString *)userID fromServer:(NSString *)pgpKeyServer error:(int *)errorCode
{
    char keyID_query[64];
    MiscINETAddress *serverAddress;
    struct in_addr address;
    MiscINETSocket *keyServer;
    NSMutableData *pgpKey;
    char keyBuf[BUFSIZE];
    NSString *keyString;
    int irep,buflen,retcode;
    NSRange noMatchRange;

    sprintf(keyID_query,"GET /pks/lookup\?op=get&search=%s\r\n",[userID cString]);
    fprintf(stderr,"PGPserver: Query String: %s\n",keyID_query);

    /* One PGP Key server is at pgp5.ai.mit.edu */
    if(!(serverAddress = [MiscINETAddress addressFromHostname:pgpKeyServer])) {
        fprintf(stderr,"PGPserver: Can't get server Address\n");
        return nil;
    }

    address = [serverAddress address];

#if 0
    fprintf(stderr,"PGPserver: Address is: %d.%d.%d.%d\n",address.S_un.S_un_b.s_b1,
            address.S_un.S_un_b.s_b2,
            address.S_un.S_un_b.s_b3,
            address.S_un.S_un_b.s_b4);
#endif

    keyServer = [[MiscINETSocket alloc] init];

    if(![keyServer connectTo:serverAddress port:11371 type:MiscSOCK_STREAM]) {
        fprintf(stderr,"PGPserver: Can't connect to port\n");
        return nil;
    }


    fprintf(stderr,"PGPserver: Sending Query\n");

    [keyServer sendData:keyID_query length:strlen(keyID_query)];
    pgpKey = nil;

    fprintf(stderr,"PGPserver: Waiting for reply\n");

    irep=0;
    do {
        buflen = BUFSIZE;
        retcode = [keyServer receiveData:keyBuf length:&buflen timeout:TIMEOUT toNext:0];
        if(!pgpKey)
          pgpKey = [NSMutableData dataWithBytes:keyBuf length:buflen];
        else
          [pgpKey appendBytes:keyBuf length:buflen];
    } while(retcode > 0);

    if(retcode==0) {
        fprintf(stderr,"PGPserver: Connection timed out\n");
        [keyString release];
        return nil;
    }

    keyString = [[NSString alloc] initWithData:pgpKey encoding:NSASCIIStringEncoding];

    noMatchRange = [keyString rangeOfString:@"No matching keys"];
    if(noMatchRange.length != 0) {
        fprintf(stderr,"PGPserver: No Matching PGP key found\n");
        [keyString release];
        return nil;
    }

    fprintf(stderr,"PGPserver: PGP Key found: %s\n",[keyString cString]);

    return [keyString autorelease];
}



@end

	     
